@@ -37,6 +37,11 @@
 #ifdef HAVE_TIME_H
 #include <time.h>
 #endif
+#ifdef HAVE_LIBUTEMPTER
+#include <utempter.h>
+#endif
+
+#include <sys/wait.h>
 
 #include <libxfce4ui/libxfce4ui.h>
 
@@ -68,6 +73,7 @@ enum
 {
   GET_CONTEXT_MENU,
   SELECTION_CHANGED,
+  CLOSE_TAB,
   LAST_SIGNAL
 };
 
@@ -119,10 +125,11 @@ static void       terminal_screen_update_misc_rewrap_on_resize  (TerminalScreen
 static void       terminal_screen_update_scrolling_lines        (TerminalScreen        *screen);
 static void       terminal_screen_update_scrolling_on_output    (TerminalScreen        *screen);
 static void       terminal_screen_update_scrolling_on_keystroke (TerminalScreen        *screen);
+static void       terminal_screen_update_text_blink_mode        (TerminalScreen        *screen);
 static void       terminal_screen_update_title                  (TerminalScreen        *screen);
 static void       terminal_screen_update_word_chars             (TerminalScreen        *screen);
 static void       terminal_screen_vte_child_exited              (VteTerminal           *terminal,
-                                                                 gint                   arg1,
+                                                                 gint                   status,
                                                                  TerminalScreen        *screen);
 static void       terminal_screen_vte_eof                       (VteTerminal           *terminal,
                                                                  TerminalScreen        *screen);
@@ -176,6 +183,9 @@ struct _TerminalScreen
 
   TerminalTitle        dynamic_title_mode;
   guint                hold : 1;
+#if !VTE_CHECK_VERSION (0, 51, 1)
+  guint                scroll_on_output : 1;
+#endif
 
   guint                activity_timeout_id;
   time_t               activity_resize_time;
@@ -250,6 +260,17 @@ terminal_screen_class_init (TerminalScreenClass *klass)
                   0, NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
+
+  /**
+   * TerminalScreen::close-tab-request
+   **/
+  screen_signals[CLOSE_TAB] =
+    g_signal_new (I_("close-tab-request"),
+                  G_TYPE_FROM_CLASS (gobject_class),
+                  G_SIGNAL_RUN_LAST,
+                  0, NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
 }
 
 
@@ -304,6 +325,7 @@ terminal_screen_init (TerminalScreen *screen)
   terminal_screen_update_scrolling_lines (screen);
   terminal_screen_update_scrolling_on_output (screen);
   terminal_screen_update_scrolling_on_keystroke (screen);
+  terminal_screen_update_text_blink_mode (screen);
   terminal_screen_update_word_chars (screen);
   terminal_screen_update_background (screen);
   terminal_screen_update_colors (screen);
@@ -347,10 +369,10 @@ terminal_screen_finalize (GObject *object)
 
 
 static void
-terminal_screen_get_property (GObject          *object,
-                              guint             prop_id,
-                              GValue           *value,
-                              GParamSpec       *pspec)
+terminal_screen_get_property (GObject    *object,
+                              guint       prop_id,
+                              GValue     *value,
+                              GParamSpec *pspec)
 {
   TerminalScreen *screen = TERMINAL_SCREEN (object);
   const gchar    *title = NULL;
@@ -417,10 +439,10 @@ terminal_screen_get_property (GObject          *object,
 
 
 static void
-terminal_screen_set_property (GObject          *object,
-                              guint             prop_id,
-                              const GValue     *value,
-                              GParamSpec       *pspec)
+terminal_screen_set_property (GObject      *object,
+                              guint         prop_id,
+                              const GValue *value,
+                              GParamSpec   *pspec)
 {
   TerminalScreen *screen = TERMINAL_SCREEN (object);
 
@@ -478,12 +500,12 @@ terminal_screen_draw (GtkWidget *widget,
                       cairo_t   *cr,
                       gpointer   user_data)
 {
-  TerminalScreen      *screen = TERMINAL_SCREEN (user_data);
-  TerminalBackground   background_mode;
-  GdkPixbuf           *image;
-  gint                 width, height;
-  cairo_surface_t     *surface;
-  cairo_t             *ctx;
+  TerminalScreen     *screen = TERMINAL_SCREEN (user_data);
+  TerminalBackground  background_mode;
+  GdkPixbuf          *image;
+  gint                width, height;
+  cairo_surface_t    *surface;
+  cairo_t            *ctx;
 
   terminal_return_val_if_fail (TERMINAL_IS_SCREEN (screen), FALSE);
   terminal_return_val_if_fail (VTE_IS_TERMINAL (screen->terminal), FALSE);
@@ -557,6 +579,10 @@ terminal_screen_preferences_changed (TerminalPreferences *preferences,
     terminal_screen_update_binding_delete (screen);
   else if (strcmp ("binding-ambiguous-width", name) == 0)
     terminal_screen_update_binding_ambiguous_width (screen);
+#if VTE_CHECK_VERSION (0, 51, 3)
+  else if (strcmp ("cell-width-scale", name) == 0 || strcmp ("cell-height-scale", name) == 0)
+    terminal_screen_update_font (screen);
+#endif
   else if (strncmp ("color-", name, strlen ("color-")) == 0)
     terminal_screen_update_colors (screen);
   else if (strncmp ("font-", name, strlen ("font-")) == 0)
@@ -579,6 +605,8 @@ terminal_screen_preferences_changed (TerminalPreferences *preferences,
     terminal_screen_update_scrolling_on_output (screen);
   else if (strcmp ("scrolling-on-keystroke", name) == 0)
     terminal_screen_update_scrolling_on_keystroke (screen);
+  else if (strcmp ("text-blink-mode", name) == 0)
+    terminal_screen_update_text_blink_mode (screen);
   else if (strncmp ("title-", name, strlen ("title-")) == 0)
     terminal_screen_update_title (screen);
   else if (strcmp ("word-chars", name) == 0)
@@ -814,6 +842,16 @@ terminal_screen_get_child_environment (TerminalScreen *screen)
           || strcmp (*p, "TERM") == 0)
         continue;
 
+#if !VTE_CHECK_VERSION (0, 51, 90)
+      /* copy working directory to $PWD, to preserve symlinks
+       * see https://bugzilla.gnome.org/show_bug.cgi?id=758452 */
+      if (strcmp (*p, "PWD") == 0)
+        {
+          result[n++] = g_strconcat (*p, "=", screen->working_directory, NULL);
+          continue;
+        }
+#endif
+
       /* copy the variable */
       value = g_getenv (*p);
       if (G_LIKELY (value != NULL))
@@ -850,8 +888,8 @@ terminal_screen_update_background (TerminalScreen *screen)
   TerminalBackground background_mode;
   gdouble            background_alpha;
 
-  terminal_return_val_if_fail (TERMINAL_IS_SCREEN (screen), FALSE);
-  terminal_return_val_if_fail (VTE_IS_TERMINAL (screen->terminal), FALSE);
+  terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
+  terminal_return_if_fail (VTE_IS_TERMINAL (screen->terminal));
 
   g_object_get (G_OBJECT (screen->preferences), "background-mode", &background_mode, NULL);
 
@@ -918,7 +956,7 @@ terminal_screen_update_binding_delete (TerminalScreen *screen)
 
   g_object_get (G_OBJECT (screen->preferences), "binding-delete", &binding, NULL);
   vte_terminal_set_delete_binding (VTE_TERMINAL (screen->terminal),
-      terminal_screen_binding_vte (binding));
+                                   terminal_screen_binding_vte (binding));
 }
 
 
@@ -930,7 +968,7 @@ terminal_screen_update_binding_ambiguous_width (TerminalScreen *screen)
 
   g_object_get (G_OBJECT (screen->preferences), "binding-ambiguous-width", &binding, NULL);
   vte_terminal_set_cjk_ambiguous_width (VTE_TERMINAL (screen->terminal),
-      binding == TERMINAL_AMBIGUOUS_WIDTH_BINDING_NARROW ? 1 : 2);
+                                        binding == TERMINAL_AMBIGUOUS_WIDTH_BINDING_NARROW ? 1 : 2);
 }
 
 
@@ -968,6 +1006,7 @@ terminal_screen_update_colors (TerminalScreen *screen)
   gboolean   vary_bg;
   gdouble    hsv[N_HSV];
   gdouble    sat_min, sat_max;
+  gboolean   bold_is_bright;
 
   g_object_get (screen->preferences,
                 "color-palette", &palette_str,
@@ -975,6 +1014,7 @@ terminal_screen_update_colors (TerminalScreen *screen)
                 "color-selection-use-default", &selection_use_default,
                 "color-bold-use-default", &bold_use_default,
                 "color-background-vary", &vary_bg,
+                "color-bold-is-bright", &bold_is_bright,
                 NULL);
 
   if (G_LIKELY (palette_str != NULL))
@@ -1051,7 +1091,7 @@ terminal_screen_update_colors (TerminalScreen *screen)
   /* cursor color */
   if (!cursor_use_default)
     {
-      cursor_use_default = !terminal_preferences_get_color (screen->preferences, "color-cursor-fg", &cursor);
+      cursor_use_default = !terminal_preferences_get_color (screen->preferences, "color-cursor-foreground", &cursor);
 #if VTE_CHECK_VERSION (0, 44, 0)
       vte_terminal_set_color_cursor_foreground (VTE_TERMINAL (screen->terminal), cursor_use_default ? NULL : &cursor);
 #endif
@@ -1064,7 +1104,7 @@ terminal_screen_update_colors (TerminalScreen *screen)
     {
       selection_use_default = !terminal_preferences_get_color (screen->preferences, "color-selection", &selection);
       vte_terminal_set_color_highlight_foreground (VTE_TERMINAL (screen->terminal), selection_use_default ? NULL : &selection);
-      selection_use_default = !terminal_preferences_get_color (screen->preferences, "color-selection-bg", &selection);
+      selection_use_default = !terminal_preferences_get_color (screen->preferences, "color-selection-background", &selection);
       vte_terminal_set_color_highlight (VTE_TERMINAL (screen->terminal), selection_use_default ? NULL : &selection);
     }
 
@@ -1073,6 +1113,11 @@ terminal_screen_update_colors (TerminalScreen *screen)
     bold_use_default = !terminal_preferences_get_color (screen->preferences, "color-bold", &bold);
   if (!bold_use_default || has_fg)
     vte_terminal_set_color_bold (VTE_TERMINAL (screen->terminal), bold_use_default ? &fg : &bold);
+
+#if VTE_CHECK_VERSION (0, 51, 3)
+  /* "bold-is-bright" supported since vte 0.51.3 */
+  vte_terminal_set_bold_is_bright (VTE_TERMINAL (screen->terminal), bold_is_bright);
+#endif
 }
 
 
@@ -1102,8 +1147,8 @@ terminal_screen_update_misc_cursor_blinks (TerminalScreen *screen)
 static void
 terminal_screen_update_misc_cursor_shape (TerminalScreen *screen)
 {
-  TerminalCursorShape    val;
-  VteCursorShape shape = VTE_CURSOR_SHAPE_BLOCK;
+  TerminalCursorShape val;
+  VteCursorShape      shape = VTE_CURSOR_SHAPE_BLOCK;
 
   g_object_get (G_OBJECT (screen->preferences), "misc-cursor-shape", &val, NULL);
 
@@ -1153,8 +1198,10 @@ terminal_screen_update_scrolling_lines (TerminalScreen *screen)
 {
   guint    lines;
   gboolean unlimited;
-  g_object_get (G_OBJECT (screen->preferences), "scrolling-lines", &lines, NULL);
-  g_object_get (G_OBJECT (screen->preferences), "scrolling-unlimited", &unlimited, NULL);
+  g_object_get (G_OBJECT (screen->preferences),
+                "scrolling-lines", &lines,
+                "scrolling-unlimited", &unlimited,
+                NULL);
   vte_terminal_set_scrollback_lines (VTE_TERMINAL (screen->terminal),
                                      unlimited ? -1 : (glong) lines);
 }
@@ -1166,7 +1213,7 @@ terminal_screen_update_scrolling_on_output (TerminalScreen *screen)
 {
   gboolean scroll;
   g_object_get (G_OBJECT (screen->preferences), "scrolling-on-output", &scroll, NULL);
-  vte_terminal_set_scroll_on_output (VTE_TERMINAL (screen->terminal), scroll);
+  terminal_screen_set_scroll_on_output (screen, scroll);
 }
 
 
@@ -1181,6 +1228,42 @@ terminal_screen_update_scrolling_on_keystroke (TerminalScreen *screen)
 
 
 
+static void
+terminal_screen_update_text_blink_mode (TerminalScreen *screen)
+{
+#if VTE_CHECK_VERSION (0, 51, 3)
+  TerminalTextBlinkMode val;
+  VteTextBlinkMode      mode = VTE_TEXT_BLINK_ALWAYS;
+
+  g_object_get (G_OBJECT (screen->preferences), "text-blink-mode", &val, NULL);
+
+  switch (val)
+    {
+      case TERMINAL_TEXT_BLINK_MODE_ALWAYS:
+        break;
+
+      case TERMINAL_TEXT_BLINK_MODE_NEVER:
+        mode = VTE_TEXT_BLINK_NEVER;
+        break;
+
+      case TERMINAL_TEXT_BLINK_MODE_FOCUSED:
+        mode = VTE_TEXT_BLINK_FOCUSED;
+        break;
+
+      case TERMINAL_TEXT_BLINK_MODE_UNFOCUSED:
+        mode = VTE_TEXT_BLINK_UNFOCUSED;
+        break;
+
+      default:
+        terminal_assert_not_reached ();
+    }
+
+  vte_terminal_set_text_blink_mode (VTE_TERMINAL (screen->terminal), mode);
+#endif
+}
+
+
+
 static void
 terminal_screen_update_title (TerminalScreen *screen)
 {
@@ -1212,14 +1295,52 @@ terminal_screen_update_word_chars (TerminalScreen *screen)
 
 static void
 terminal_screen_vte_child_exited (VteTerminal    *terminal,
-                                  gint            arg1,
+                                  gint            status,
                                   TerminalScreen *screen)
 {
+  GtkWidget *window, *dialog, *label, *content_area;
+  gchar     *message;
+  gint       response;
+
   terminal_return_if_fail (VTE_IS_TERMINAL (terminal));
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
 
   if (G_LIKELY (!screen->hold))
     gtk_widget_destroy (GTK_WIDGET (screen));
+  else
+    {
+      /* create "Relaunch" dialog */
+      window = gtk_widget_get_toplevel (GTK_WIDGET (screen));
+      dialog = gtk_dialog_new_with_buttons (_("Child process exited"),
+                                            GTK_WINDOW (window),
+                                            GTK_DIALOG_DESTROY_WITH_PARENT,
+                                            _("_Relaunch"),
+                                            GTK_RESPONSE_YES,
+                                            NULL);
+      gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_YES);
+
+      if (WIFEXITED (status))
+        message = g_strdup_printf (_("The child process exited normally with status %d."), WEXITSTATUS (status));
+      else if (WIFSIGNALED (status))
+        message = g_strdup_printf (_("The child process was aborted by signal %d."), WTERMSIG (status));
+      else
+        message = g_strdup_printf (_("The child process was aborted."));
+
+      label = gtk_label_new (message);
+      g_free (message);
+
+      content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
+      gtk_container_add (GTK_CONTAINER (content_area), label);
+
+      gtk_widget_show_all (dialog);
+      response = gtk_dialog_run (GTK_DIALOG (dialog));
+
+      /* relaunch the process */
+      if (response == GTK_RESPONSE_YES)
+        terminal_screen_launch_child (screen);
+
+      gtk_widget_destroy (dialog);
+    }
 }
 
 
@@ -1261,7 +1382,7 @@ terminal_screen_vte_selection_changed (VteTerminal    *terminal,
   g_object_get (G_OBJECT (screen->preferences),
                 "misc-copy-on-select", &copy_on_select, NULL);
   if (copy_on_select && vte_terminal_get_has_selection (terminal))
-    vte_terminal_copy_clipboard (terminal);
+    terminal_screen_copy_clipboard (screen);
 
   g_signal_emit (G_OBJECT (screen), screen_signals[SELECTION_CHANGED], 0);
 }
@@ -1464,8 +1585,8 @@ terminal_screen_update_label_orientation (TerminalScreen *screen)
 
 
 static gchar*
-terminal_screen_zoom_font (TerminalScreen *screen,
-                           gchar *font_name,
+terminal_screen_zoom_font (TerminalScreen   *screen,
+                           gchar            *font_name,
                            TerminalZoomLevel zoom)
 {
   gdouble               scale;
@@ -1510,9 +1631,9 @@ terminal_screen_zoom_font (TerminalScreen *screen,
   pango_font_description_free (font_desc);
 
   if (font_zoomed == NULL)
-      return font_name;
+    return font_name;
 
-  g_free(font_name);
+  g_free (font_name);
 
   return font_zoomed;
 }
@@ -1523,18 +1644,14 @@ static void
 terminal_screen_urgent_bell (TerminalWidget *widget,
                              TerminalScreen *screen)
 {
-  GtkWidget *toplevel;
-  gboolean   enabled;
+  gboolean enabled;
 
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
 
   g_object_get (G_OBJECT (screen->preferences), "misc-bell-urgent", &enabled, NULL);
 
-  if (!enabled)
-    return;
-
-  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (screen));
-  gtk_window_set_urgency_hint (GTK_WINDOW (toplevel), TRUE);
+  if (enabled)
+    gtk_window_set_urgency_hint (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (screen))), TRUE);
 }
 
 
@@ -1605,6 +1722,9 @@ terminal_screen_launch_child (TerminalScreen *screen)
   guint         i;
   VtePtyFlags   pty_flags = VTE_PTY_DEFAULT;
   GSpawnFlags   spawn_flags = G_SPAWN_CHILD_INHERITS_STDIN | G_SPAWN_SEARCH_PATH;
+#ifdef HAVE_LIBUTEMPTER
+  gboolean      update_records;
+#endif
 
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
 
@@ -1647,6 +1767,12 @@ terminal_screen_launch_child (TerminalScreen *screen)
           g_error_free (error);
         }
 
+#ifdef HAVE_LIBUTEMPTER
+      g_object_get (G_OBJECT (screen->preferences), "command-update-records", &update_records, NULL);
+      if (update_records)
+        utempter_add_record (vte_pty_get_fd (vte_terminal_get_pty (VTE_TERMINAL (screen->terminal))), NULL);
+#endif
+
       g_free (argv2);
 
       g_strfreev (argv);
@@ -1787,8 +1913,8 @@ terminal_screen_set_window_geometry_hints (TerminalScreen *screen,
 
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
   terminal_return_if_fail (VTE_IS_TERMINAL (screen->terminal));
-  terminal_return_if_fail (gtk_widget_get_realized (screen));
-  terminal_return_if_fail (gtk_widget_get_realized (window));
+  terminal_return_if_fail (gtk_widget_get_realized (GTK_WIDGET (screen)));
+  terminal_return_if_fail (gtk_widget_get_realized (GTK_WIDGET (window)));
 
 #if GTK_CHECK_VERSION (3, 19, 5)
   terminal_screen_get_geometry (screen, &char_width, &char_height, NULL, NULL);
@@ -1851,7 +1977,6 @@ terminal_screen_force_resize_window (TerminalScreen *screen,
   glong          char_width;
   glong          char_height;
 
-
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
   terminal_return_if_fail (VTE_IS_TERMINAL (screen->terminal));
   terminal_return_if_fail (GTK_IS_WINDOW (window));
@@ -1977,14 +2102,22 @@ terminal_screen_get_title (TerminalScreen *screen)
 const gchar*
 terminal_screen_get_working_directory (TerminalScreen *screen)
 {
-  gchar  buffer[4096 + 1];
-  gchar *file;
-  gchar *cwd;
-  gint   length;
+  gchar        buffer[4096 + 1];
+  gchar       *file;
+  gchar       *cwd;
+  const gchar *uri;
+  gint         length;
 
   terminal_return_val_if_fail (TERMINAL_IS_SCREEN (screen), NULL);
 
-  if (screen->pid >= 0)
+  /* try to use vte functionality first: see bug #13902 */
+  uri = vte_terminal_get_current_directory_uri (VTE_TERMINAL (screen->terminal));
+  if (uri != NULL)
+    {
+      g_free (screen->working_directory);
+      screen->working_directory = g_filename_from_uri (uri, NULL, NULL);
+    }
+  else if (screen->pid >= 0)
     {
       /* make sure that we use linprocfs on all systems */
 #if defined(__FreeBSD__)
@@ -2066,17 +2199,39 @@ terminal_screen_has_selection (TerminalScreen *screen)
  * terminal_screen_copy_clipboard:
  * @screen  : A #TerminalScreen.
  *
- * Places the selected text in the terminal in the #GDK_SELECTIN_CLIPBOARD selection.
+ * Places the selected text in the terminal in the #GDK_SELECTION_CLIPBOARD selection.
  **/
 void
 terminal_screen_copy_clipboard (TerminalScreen *screen)
 {
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
+#if VTE_CHECK_VERSION (0, 49, 2)
+  vte_terminal_copy_clipboard_format (VTE_TERMINAL (screen->terminal), VTE_FORMAT_TEXT);
+#else
   vte_terminal_copy_clipboard (VTE_TERMINAL (screen->terminal));
+#endif
 }
 
 
 
+/**
+ * terminal_screen_copy_clipboard_html:
+ * @screen  : A #TerminalScreen.
+ *
+ * Places the selected text in the terminal in the #GDK_SELECTION_CLIPBOARD selection
+ * as HTML (preserving colors, bold font, etc).
+ **/
+#if VTE_CHECK_VERSION (0, 49, 2)
+void
+terminal_screen_copy_clipboard_html (TerminalScreen *screen)
+{
+  terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
+  vte_terminal_copy_clipboard_format (VTE_TERMINAL (screen->terminal), VTE_FORMAT_HTML);
+}
+#endif
+
+
+
 /**
  * terminal_screen_paste_clipboard:
  * @screen  : A #TerminalScreen.
@@ -2202,6 +2357,14 @@ terminal_screen_reset_activity (TerminalScreen *screen)
 
 
 
+static void
+terminal_screen_close_tab_cb (TerminalScreen *screen)
+{
+  g_signal_emit (G_OBJECT (screen), screen_signals[CLOSE_TAB], 0);
+}
+
+
+
 GtkWidget *
 terminal_screen_get_tab_label (TerminalScreen *screen)
 {
@@ -2238,7 +2401,7 @@ terminal_screen_get_tab_label (TerminalScreen *screen)
   gtk_widget_set_valign (button, GTK_ALIGN_CENTER);
   gtk_container_add (GTK_CONTAINER (hbox), button);
   g_signal_connect_swapped (G_OBJECT (button), "clicked",
-                            G_CALLBACK (gtk_widget_destroy), screen);
+                            G_CALLBACK (terminal_screen_close_tab_cb), screen);
 
   /* button image */
   image = gtk_image_new_from_icon_name ("window-close-symbolic", GTK_ICON_SIZE_MENU);
@@ -2264,7 +2427,6 @@ void
 terminal_screen_focus (TerminalScreen *screen)
 {
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
-
   gtk_widget_grab_focus (GTK_WIDGET (screen->terminal));
 }
 
@@ -2286,9 +2448,8 @@ terminal_screen_set_encoding (TerminalScreen *screen,
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
   if (charset == NULL)
     g_get_charset (&charset);
-  if (!vte_terminal_set_encoding (VTE_TERMINAL (screen->terminal), charset, NULL)) {
+  if (!vte_terminal_set_encoding (VTE_TERMINAL (screen->terminal), charset, NULL))
     g_printerr (_("Failed to set encoding %s\n"), charset);
-  }
 }
 
 
@@ -2395,6 +2556,9 @@ terminal_screen_update_font (TerminalScreen *screen)
   PangoFontDescription *font_desc;
   glong                 grid_w = 0, grid_h = 0;
   GSettings            *settings;
+#if VTE_CHECK_VERSION (0, 51, 3)
+  gdouble cell_width_scale, cell_height_scale;
+#endif
 
   terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
   terminal_return_if_fail (TERMINAL_IS_PREFERENCES (screen->preferences));
@@ -2416,14 +2580,15 @@ terminal_screen_update_font (TerminalScreen *screen)
 
   if (TERMINAL_IS_WINDOW (toplevel))
     {
-      if (TERMINAL_WINDOW (toplevel)->font)
+      if (terminal_window_get_font (TERMINAL_WINDOW (toplevel)))
         {
           g_free (font_name);
-          font_name = g_strdup (TERMINAL_WINDOW (toplevel)->font);
+          font_name = g_strdup (terminal_window_get_font (TERMINAL_WINDOW (toplevel)));
         }
 
-      if (TERMINAL_WINDOW (toplevel)->zoom != TERMINAL_ZOOM_LEVEL_DEFAULT)
-        font_name = terminal_screen_zoom_font (screen, font_name, TERMINAL_WINDOW (toplevel)->zoom);
+      if (terminal_window_get_zoom_level (TERMINAL_WINDOW (toplevel)) != TERMINAL_ZOOM_LEVEL_DEFAULT)
+        font_name = terminal_screen_zoom_font (screen, font_name,
+                                               terminal_window_get_zoom_level (TERMINAL_WINDOW (toplevel)));
     }
 
   if (gtk_widget_get_realized (GTK_WIDGET (screen)))
@@ -2432,12 +2597,25 @@ terminal_screen_update_font (TerminalScreen *screen)
   if (G_LIKELY (font_name != NULL))
     {
       font_desc = pango_font_description_from_string (font_name);
+#if !VTE_CHECK_VERSION (0, 51, 3)
+      /* "bold-is-bright" is used since vte 0.51.3 */
       vte_terminal_set_allow_bold (VTE_TERMINAL (screen->terminal), font_allow_bold);
+#endif
       vte_terminal_set_font (VTE_TERMINAL (screen->terminal), font_desc);
       pango_font_description_free (font_desc);
       g_free (font_name);
     }
 
+#if VTE_CHECK_VERSION (0, 51, 3)
+  g_object_get (G_OBJECT (screen->preferences),
+                "cell-width-scale", &cell_width_scale,
+                "cell-height-scale", &cell_height_scale,
+                NULL);
+
+  vte_terminal_set_cell_width_scale (VTE_TERMINAL (screen->terminal), cell_width_scale);
+  vte_terminal_set_cell_height_scale (VTE_TERMINAL (screen->terminal), cell_height_scale);
+#endif
+
   /* update window geometry it required (not needed for drop-down) */
   if (TERMINAL_IS_WINDOW (toplevel) && !TERMINAL_WINDOW (toplevel)->drop_down && grid_w > 0 && grid_h > 0)
     terminal_screen_force_resize_window (screen, GTK_WINDOW (toplevel), grid_w, grid_h);
@@ -2445,6 +2623,15 @@ terminal_screen_update_font (TerminalScreen *screen)
 
 
 
+gboolean
+terminal_screen_get_input_enabled (TerminalScreen *screen)
+{
+  terminal_return_val_if_fail (TERMINAL_IS_SCREEN (screen), FALSE);
+  return vte_terminal_get_input_enabled (VTE_TERMINAL (screen->terminal));
+}
+
+
+
 void
 terminal_screen_set_input_enabled (TerminalScreen *screen,
                                    gboolean        enabled)
@@ -2455,6 +2642,32 @@ terminal_screen_set_input_enabled (TerminalScreen *screen,
 
 
 
+gboolean
+terminal_screen_get_scroll_on_output (TerminalScreen *screen)
+{
+  terminal_return_val_if_fail (TERMINAL_IS_SCREEN (screen), FALSE);
+#if !VTE_CHECK_VERSION (0, 51, 1)
+  return screen->scroll_on_output;
+#else
+  return vte_terminal_get_scroll_on_output (VTE_TERMINAL (screen->terminal));
+#endif
+}
+
+
+
+void
+terminal_screen_set_scroll_on_output (TerminalScreen *screen,
+                                      gboolean        enabled)
+{
+  terminal_return_if_fail (TERMINAL_IS_SCREEN (screen));
+#if !VTE_CHECK_VERSION (0, 51, 1)
+  screen->scroll_on_output = enabled;
+#endif
+  vte_terminal_set_scroll_on_output (VTE_TERMINAL (screen->terminal), enabled);
+}
+
+
+
 void
 terminal_screen_save_contents (TerminalScreen *screen,
                                GOutputStream  *stream,
@@ -2464,3 +2677,36 @@ terminal_screen_save_contents (TerminalScreen *screen,
   vte_terminal_write_contents_sync (VTE_TERMINAL (screen->terminal),
                                     stream, VTE_WRITE_DEFAULT, NULL, &error);
 }
+
+
+
+/**
+ * terminal_screen_has_foreground_process:
+ * @screen  : A #TerminalScreen.
+ *
+ * Return value: %TRUE if there's a foreground process running in @screen.
+ **/
+gboolean
+terminal_screen_has_foreground_process (TerminalScreen *screen)
+{
+  VtePty *pty;
+  int     fd;
+  int     fgpid;
+
+  if (screen == NULL)
+    return FALSE;
+
+  pty = vte_terminal_get_pty (VTE_TERMINAL (screen->terminal));
+  if (pty == NULL)
+    return FALSE;
+
+  fd = vte_pty_get_fd (pty);
+  if (fd == -1)
+    return FALSE;
+
+  fgpid = tcgetpgrp (fd);
+  if (fgpid == -1 || fgpid == screen->pid)
+    return FALSE;
+
+  return TRUE;
+}
